/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package engcalculator.function.prefix;

import engcalculator.function.prefix.square.*;
import engcalculator.function.prefix.interval.*;
import engcalculator.function.prefix.system.*;
import engcalculator.function.prefix.calculus.*;
import engcalculator.function.prefix.cumulative.*;
import engcalculator.function.prefix.logarithm.*;
import engcalculator.function.*;
import engcalculator.function.prefix.information.*;
import engcalculator.function.prefix.engineering.*;
import engcalculator.function.prefix.equation.*;
import engcalculator.function.prefix.plot.*;
import engcalculator.function.prefix.flowControl.*;
import engcalculator.function.prefix.statistic.*;
import engcalculator.function.prefix.decibel.*;
import engcalculator.function.prefix.space.*;
import engcalculator.function.prefix.matrix.*;
import engcalculator.function.prefix.trigonometry.*;
import engcalculator.data.IntervalList;
import engcalculator.data.MatrixIntervalList;
import engcalculator.expression.Expression;
import engcalculator.expression.token.InfixFunctionTkn;
import engcalculator.expression.token.PrefixFunctionTkn;
import java.util.Collection;
import java.util.HashMap;

/**
 *
 * @author simone
 */
public abstract class PrefixFunction extends Function {
    private static HashMap<String, PrefixFunction> PREFIX_FUNCTIONS = new HashMap <String, PrefixFunction> ();
    private static final PrefixFunctionTkn PREFIX_FUNCTION_TKN = new PrefixFunctionTkn();
    private static final InfixFunctionTkn INFIX_FUNCTION_TKN = new InfixFunctionTkn();
    private static final double MAX_CHECK_ERROR = 1e-3;
    private final static boolean DEBUG = false;


    public final static int findCloserIndex (final float[] values, float value, boolean ascendingSorted) {
        int closerIndex = 0;
        float minDistance = Math.abs(value - values[0]), distance;
        for (int index = 1; index < values.length; ++index) {
            distance = Math.abs(value - values[index]);
            if (distance < minDistance) {
                closerIndex = index;
                minDistance = distance;
            } else  if (ascendingSorted) break; //it would be faster to use binary search for ascending ordered array
        }
        return closerIndex;
    }

    public final static int findCloserIndex (final int[] values, int value, boolean ascendingSorted) {
        int closerIndex = 0;
        int minDistance = Math.abs(value - values[0]), distance;
        for (int index = 1; index < values.length; ++index) {
            distance = Math.abs(value - values[index]);
            if (distance < minDistance) {
                closerIndex = index;
                minDistance = distance;
            } else if (ascendingSorted) break; //it would be faster to use binary search for ascending ordered array
        }
        return closerIndex;

    }

    public final static int findCloserIndex (final String[] values, String value, boolean ascendingSorted) {
        int closerIndex = 0;
        int minDistance = Math.abs(value.compareTo(values[0])), distance;
        for (int index = 1; index < values.length; ++index) {
            distance = Math.abs(value.compareTo(values[index]));
            if (distance < minDistance) {
                closerIndex = index;
                minDistance = distance;
            } else if (ascendingSorted) break; //it would be faster to use binary search for ascending ordered array
        }
        return closerIndex;

    }

    public static boolean addFunction (PrefixFunction func) throws Exception {
        
        if(!((func.isConstant() && INFIX_FUNCTION_TKN.mayBeCompatible(func.getSymbol())) || PREFIX_FUNCTION_TKN.mayBeCompatible(func.getSymbol()))) throw new CalculusException("The prefix function name is not properly formed: "+func.getSymbol());

        Function f;

        if ((f = PREFIX_FUNCTIONS.get(func.getSymbol())) != null && f.isConstant()) return false;

        PREFIX_FUNCTIONS.put(func.getSymbol(), func);

        return true;
    }

    public static boolean delFunction(String value) {
        Function f;
        if ((f = PREFIX_FUNCTIONS.get(value)) != null && ! f.isConstant()) {
            PREFIX_FUNCTIONS.remove(value);
            return true;
        }
        return false;
    }

    public static PrefixFunction getFunction (String symbol) {
        return PREFIX_FUNCTIONS.get(symbol);
    }

    public static Collection<PrefixFunction> getPrefixFunctions () {
        return PREFIX_FUNCTIONS.values();
    }

    private static String buildHelp (String name, boolean expandsToList, String basehelp, String[] examples, String[] results) {
        StringBuffer help = new StringBuffer (basehelp);
        if (expandsToList) help.append(" If the argument is a list the function is automatically applied to all the elements of the list providing as result the list of the results.");
        if (examples != null && results != null && examples.length == results.length)  {
            help.append("\nExamples: ");
            for (int i=0; i<examples.length; ++i) {
                if( examples[i].indexOf(';') == -1) help.append ('\n').append(name).append('(').append(examples[i]).append(')');
                else help.append ('\n').append(examples[i]);
                help.append(" = ").append(results[i]);
            }
        }
        return help.toString();
    }

    public static void initialize () throws Exception {
        boolean noDuplicateName = true;

        noDuplicateName &= addFunction (new Dual());
        noDuplicateName &= addFunction (new Direction());
        noDuplicateName &= addFunction (new Sign());
        noDuplicateName &= addFunction (new Proper());
        noDuplicateName &= addFunction (new Distance());
        noDuplicateName &= addFunction (new Abs());
        noDuplicateName &= addFunction (new Module());

        noDuplicateName &= addFunction(new Sin());
        noDuplicateName &= addFunction(new Cos());
        noDuplicateName &= addFunction(new Tan());
        noDuplicateName &= addFunction(new ArcCos());
        noDuplicateName &= addFunction(new ArcSin());
        noDuplicateName &= addFunction(new ArcTan());

        noDuplicateName &= addFunction(new Extremes());
        noDuplicateName &= addFunction(new ValTol());

        noDuplicateName &= addFunction(new Square());
        noDuplicateName &= addFunction(new SquareRoot());

        noDuplicateName &= addFunction(new Ln());
        noDuplicateName &= addFunction(new Log());

        noDuplicateName &= addFunction(new IncSpace());
        noDuplicateName &= addFunction(new IntSpace());
        noDuplicateName &= addFunction(new LinSpace());
        noDuplicateName &= addFunction(new LogSpace());

        noDuplicateName &= addFunction(new PreferredValue());
        noDuplicateName &= addFunction(new ToColorCode());
        noDuplicateName &= addFunction(new FromColorCode());

        noDuplicateName &= addFunction(new Delete());
        noDuplicateName &= addFunction(new Help());
        noDuplicateName &= addFunction(new Quit());
        noDuplicateName &= addFunction(new PrintMode());
        noDuplicateName &= addFunction(new OuterExtimation());

        noDuplicateName &= addFunction(new Negate());
        noDuplicateName &= addFunction(new Invert());
        noDuplicateName &= addFunction(new DoNothing());

        noDuplicateName &= addFunction(new SumUp());
        noDuplicateName &= addFunction(new MulUp());
        noDuplicateName &= addFunction(new SumSquaredUp());
        noDuplicateName &= addFunction(new SumInvertedUp());


        noDuplicateName &= addFunction(new Derivate ());
        noDuplicateName &= addFunction(new Integrate());

        noDuplicateName &= addFunction(new Partition());

        noDuplicateName &= addFunction(new DBAmplitudeToLin());
        noDuplicateName &= addFunction(new DBPowerToLin());
        noDuplicateName &= addFunction(new LinAmplitudeToDB());
        noDuplicateName &= addFunction(new LinPowerToDB());

        noDuplicateName &= addFunction(new Average());
        noDuplicateName &= addFunction(new Size());
        noDuplicateName &= addFunction(new ColumnSize());
        noDuplicateName &= addFunction(new RowSize());
        noDuplicateName &= addFunction(new Max());
        noDuplicateName &= addFunction(new Min());
        noDuplicateName &= addFunction(new Variance());

        noDuplicateName &= addFunction (new Instruction());
        noDuplicateName &= addFunction (new About());
        noDuplicateName &= addFunction (new Version());
        noDuplicateName &= addFunction (new Examples ());

        noDuplicateName &= addFunction (new MatrixTranspose ());
        noDuplicateName &= addFunction (new MatrixIdentity ());
        noDuplicateName &= addFunction (new MatrixDiagonal ());

        noDuplicateName &= addFunction (new Rand());
        noDuplicateName &= addFunction (new TimeMs());

        noDuplicateName &= addFunction (new GaussElimination());
        noDuplicateName &= addFunction (new GaussSeidel());
        noDuplicateName &= addFunction (new FindRoot());

        noDuplicateName &= addFunction (new RepeatThisWhile());
        noDuplicateName &= addFunction (new RepeatNextWhile());
        noDuplicateName &= addFunction (new RepeatPreviousWhile());
        noDuplicateName &= addFunction (new SkipNextIf());
        noDuplicateName &= addFunction (new StopIf());

        noDuplicateName &= addFunction (new StandardPlot());
        noDuplicateName &= addFunction (new ParametricPlot());
        noDuplicateName &= addFunction (new SurfacePlot());
        
        if (! noDuplicateName) throw new CalculusException ("Duplicated Prefix Function definition while initializing internal functions.");
    }

    private byte numbArgs;          //number of arguments, <0 for variable
    private boolean expandsToLists; //if exapndible (and number of arguments is not variable) it is automatically applied to all the components of the list
    private String[] examples;
    private String[] results;

    public PrefixFunction(String name, byte numbArgs, boolean expandsToLists, boolean constant, String help, String[] examples, String[] results) {
        super(name, constant, buildHelp (name, numbArgs > 0 && expandsToLists, help, examples, results));
        this.numbArgs = numbArgs;
        this.expandsToLists = expandsToLists;
        this.examples = examples;
        this.results = results;
    }

    protected String selfCheck () {
        IntervalList computedRes, expectedRes;
        String expression=null;
        double error;

        if (examples == null && results == null ) return null;
        if (examples != null && results != null && examples.length == results.length)  {
            for (int i=0; i<examples.length; ++i) {
                try {
                    if (examples[i].indexOf(";") == -1) expression = new StringBuffer(getSymbol()).append(" (").append(examples[i]).append(')').toString();
                    else expression = examples[i];
                        
                    computedRes = new Expression (expression, null).evaluate(null);
                    
                    expectedRes = new Expression(results[i], null).evaluate(null);

                    if (computedRes.size() != expectedRes.size()) return new StringBuffer(expression).append("gives ").append(computedRes).append("which does not match in size with").append(expectedRes).toString();
                    error = 0;
                    for (int j=0; j<computedRes.size(); ++j)
                        error += Math.abs(computedRes.get(j).getRight()-expectedRes.get(j).getRight())+Math.abs(computedRes.get(j).getLeft()-expectedRes.get(j).getLeft());
                    
                } catch (Exception e) {
                    return new StringBuffer("Exception computing: ").append(e.getMessage()).append(expression).toString();
                }
                if (error > MAX_CHECK_ERROR) return new StringBuffer(expression).append(" == ").append(computedRes).append(" != ").append(results[i]).toString();
            }
        } else return new StringBuffer("Examples size for ").append(getSymbol()).append(" do not match.").toString();
        return null;
    }

    public byte getNumbArgs() {
        return numbArgs;
    }

    @Override
    public IntervalList compute (IntervalList... inputArray) throws Exception {
        if (inputArray.length != getNumberOfArguments()) throw new Exception ("It was expected just one argument for a prefix function");
        IntervalList input = inputArray[0];

        if (DEBUG) System.out.println ("Evaluating prefix function: "+getSymbol()+" with input= "+input);

        if ((numbArgs > 0 && (! expandsToLists) && input.size() != numbArgs) || (numbArgs == 0 && input.size() > 0)) throw new CalculusException ("Wrong number of arguments evaluating "+getSymbol()+" expected "+numbArgs+" received "+input.size());


        if (numbArgs > 0 && numbArgs < input.size() && expandsToLists) {
            IntervalList result;

            if (input.isMatrix()) result = new MatrixIntervalList (input.getColumns());
            else result = new IntervalList ();

            if (input.size() % numbArgs != 0) throw new CalculusException ("Wrong number of arguments evaluating "+getSymbol()+" expected "+numbArgs+" received "+input.size());

            for(int i=0;i<input.size();i+=numbArgs) {
                result.addAll( _compute( input.subList(i, i+numbArgs) ));
            }

            return result;
        } else if (numbArgs < 0 || numbArgs == input.size()) {//the followint statement lets every function called with EXACT number of argument to retrun an extension of IntervalList which may be required!!!
            return _compute( input );
        } else {
            throw new CalculusException ("Wrong number of arguments evaluating "+getSymbol()+" expected "+numbArgs+" received "+input.size());
        }
    }

    public abstract IntervalList _compute (IntervalList input) throws Exception;

    @Override
    public String getHelp() {
        if (super.getHelp().equals("")) return "";

        StringBuffer result = new StringBuffer();
        result.append("\nName: ").append(getSymbol()).append("\nType: prefix function\nNumber of args: ");
        if (numbArgs < 0) result.append("variable");
        else result.append(numbArgs);
        if (expandsToLists) result.append("\nAutomatically extendes to list");
        else result.append("\nDoes not extend to list");
        if (isConstant()) result.append("\nCannot be changed by user");
        else result.append("\nCan be changed by user");
        result.append("\nHelp: ").append(super.getHelp()).append('\n');
        return result.toString();
    }

    @Override
    public final int getNumberOfArguments() {
        return 1;
    }
}
